 // Copyright (C) 1999-2000 Id Software, Inc.
//
// bg_pmove.c -- both games player movement code
// takes a playerstate and a usercmd as input and returns a modifed playerstate

#include "q_shared.h"
#include "bg_public.h"
#include "bg_local.h"

pmove_t		*pm;
pml_t		pml;

// movement parameters
float	pm_stopspeed = 100;
float	pm_duckScale = 0.25;
float	pm_backScale = 0.65;
float	pm_swimScale = 0.50;
float	pm_wadeScale = 0.70;

float	pm_accelerate = 10;
float	pm_airaccelerate = 1;
float	pm_wateraccelerate = 4;
float	pm_flyaccelerate = 8;

float	pm_friction = 6;
float	pm_waterfriction = 1;
float	pm_flightfriction = 2;

int		c_pmove = 0;


/*
===============
PM_AddEvent

===============
*/
void PM_AddEvent( int newEvent ) {	
	BG_AddPredictableEventToPlayerstate( newEvent, 0, pm->ps );
}

/*
===============
PM_AddTouchEnt
===============
*/
void PM_AddTouchEnt( int entityNum ) {
	int		i;

	if ( entityNum == ENTITYNUM_WORLD ) {
		return;
	}
	if ( pm->numtouch == MAXTOUCH ) {
		return;
	}

	// see if it is already added
	for ( i = 0 ; i < pm->numtouch ; i++ ) {
		if ( pm->touchents[ i ] == entityNum ) {
			return;
		}
	}

	// add it
	pm->touchents[pm->numtouch] = entityNum;
	pm->numtouch++;
}

/*
===================
PM_StartTorsoAnim
===================
*/
static void PM_StartTorsoAnim( int anim ) {
	if ( pm->ps->pm_type >= PM_DEAD ) {
		return;
	}
	pm->ps->torsoAnim = ( ( pm->ps->torsoAnim & ANIM_TOGGLEBIT ) ^ ANIM_TOGGLEBIT )
		| anim;
}
static void PM_StartLegsAnim( int anim ) {
	if ( pm->ps->pm_type >= PM_DEAD ) {
		return;
	}
	if ( pm->ps->legsTimer > 0 ) {
		return;		// a high priority animation is running
	}
	pm->ps->legsAnim = ( ( pm->ps->legsAnim & (ANIM_TOGGLEBIT | ANIM_CLIMBWALLS) ) ^ ANIM_TOGGLEBIT )
		| anim;
}

static void PM_ContinueLegsAnim( int anim ) {
	if ( ( pm->ps->legsAnim & ~(ANIM_TOGGLEBIT | ANIM_CLIMBWALLS)) == anim ) {
		return;
	}
	if ( pm->ps->legsTimer > 0 ) {
		return;		// a high priority animation is running
	}
	PM_StartLegsAnim( anim );
}

static void PM_ContinueTorsoAnim( int anim ) {
	if ( ( pm->ps->torsoAnim & ~(ANIM_TOGGLEBIT | ANIM_CLIMBWALLS) ) == anim ) {
		return;
	}
	if ( pm->ps->torsoTimer > 0 ) {
		return;		// a high priority animation is running
	}
	PM_StartTorsoAnim( anim );
}

static void PM_ForceLegsAnim( int anim ) {
	pm->ps->legsTimer = 0;
	PM_StartLegsAnim( anim );
}


/*
==================
PM_ClipVelocity

Slide off of the impacting surface
==================
*/
void PM_ClipVelocity( vec3_t in, vec3_t normal, vec3_t out, float overbounce ) {
	float	backoff;
	float	change;
	int		i;
	
	backoff = DotProduct (in, normal);
	
	if ( backoff < 0 ) {
		backoff *= overbounce;
	} else {
		backoff /= overbounce;
	}

	for ( i=0 ; i<3 ; i++ ) {
		change = normal[i]*backoff;
		out[i] = in[i] - change;
	}
}


/*
==================
PM_Friction

Handles both ground friction and water friction
==================
*/
static void PM_Friction( void ) {
	vec3_t	vec;
	float	*vel;
	float	speed, newspeed, control;
	float	drop;
	
	vel = pm->ps->velocity;
	
	VectorCopy( vel, vec );
	if ( pml.walking ) {
		vec[2] = 0;	// ignore slope movement
	}

	speed = VectorLength(vec);
	if (speed < 1) {
		vel[0] = 0;
		vel[1] = 0;		// allow sinking underwater
		// FIXME: still have z friction underwater?
		return;
	}

	drop = 0;

	// apply ground friction
	if ( pm->waterlevel <= 1 ) {
		if ( pml.walking && !(pml.groundTrace.surfaceFlags & SURF_SLICK) ) {
			// if getting knocked back, no friction
			if ( ! (pm->ps->pm_flags & PMF_TIME_KNOCKBACK) ) {
				control = speed < pm_stopspeed ? pm_stopspeed : speed;
				drop += control*pm_friction*pml.frametime;
			}
		}
	}

	// apply water friction even if just wading
	if ( pm->waterlevel ) {
		drop += speed*pm_waterfriction*pm->waterlevel*pml.frametime;
	}

	// apply flying friction
	if (  pm->ps->pm_type == PM_SPECTATOR ) {
		drop += speed*pm_flightfriction*pml.frametime;
	}
	else if(pm->ps->powerups[PW_FLIGHT])
	{
		float flightfriction = pm_flightfriction;
		if(pm->cmd.forwardmove > 0 && pm->cmd.rightmove == 0 && pm->cmd.upmove <= 0)
		{
			flightfriction *= 0.2;
		}
		drop += speed*flightfriction*pml.frametime;
	}

	// scale the velocity
	newspeed = speed - drop;
	if (newspeed < 0) {
		newspeed = 0;
	}
	newspeed /= speed;

	vel[0] = vel[0] * newspeed;
	vel[1] = vel[1] * newspeed;
	vel[2] = vel[2] * newspeed;
}


/*
==============
PM_Accelerate

Handles user intended acceleration
==============
*/
static void PM_Accelerate( vec3_t wishdir, float wishspeed, float accel ) {
#if 1 
	// q2 style
	int			i;
	float		addspeed, accelspeed, currentspeed;

	currentspeed = DotProduct (pm->ps->velocity, wishdir);
	addspeed = wishspeed - currentspeed;
	if (addspeed <= 0) {
		return;
	}
	accelspeed = accel*pml.frametime*wishspeed;
	if (accelspeed > addspeed) {
		accelspeed = addspeed;
	}
	
	for (i=0 ; i<3 ; i++) {
		pm->ps->velocity[i] += accelspeed*wishdir[i];	
	}
#else
	// proper way (avoids strafe jump maxspeed bug), but feels bad
	vec3_t		wishVelocity;
	vec3_t		pushDir;
	float		pushLen;
	float		canPush;

	VectorScale( wishdir, wishspeed, wishVelocity );
	VectorSubtract( wishVelocity, pm->ps->velocity, pushDir );
	pushLen = VectorNormalize( pushDir );

	canPush = accel*pml.frametime*wishspeed;
	if (canPush > pushLen) {
		canPush = pushLen;
	}

	VectorMA( pm->ps->velocity, canPush, pushDir, pm->ps->velocity );
#endif
}



/*
============
PM_CmdScale

Returns the scale factor to apply to cmd movements
This allows the clients to use axial -127 to 127 values for all directions
without getting a sqrt(2) distortion in speed.
============
*/
static float PM_CmdScale( usercmd_t *cmd ) {
	int		max;
	float	total;
	float	scale;

	max = abs( cmd->forwardmove );
	if ( abs( cmd->rightmove ) > max ) {
		max = abs( cmd->rightmove );
	}
	if ( abs( cmd->upmove ) > max ) {
		max = abs( cmd->upmove );
	}
	if ( !max ) {
		return 0;
	}

	total = sqrt( cmd->forwardmove * cmd->forwardmove
		+ cmd->rightmove * cmd->rightmove + cmd->upmove * cmd->upmove );
	scale = (float)pm->ps->speed * max / ( 127.0 * total );
	
	return scale;
}


/*
================
PM_SetMovementDir

Determine the rotation of the legs reletive
to the facing dir
================
*/
static void PM_SetMovementDir( void ) {
	if ( pm->cmd.forwardmove || pm->cmd.rightmove ) {
		if ( pm->cmd.rightmove == 0 && pm->cmd.forwardmove > 0 ) {
			pm->ps->movementDir = 0;
		} else if ( pm->cmd.rightmove < 0 && pm->cmd.forwardmove > 0 ) {
			pm->ps->movementDir = 1;
		} else if ( pm->cmd.rightmove < 0 && pm->cmd.forwardmove == 0 ) {
			pm->ps->movementDir = 2;
		} else if ( pm->cmd.rightmove < 0 && pm->cmd.forwardmove < 0 ) {
			pm->ps->movementDir = 3;
		} else if ( pm->cmd.rightmove == 0 && pm->cmd.forwardmove < 0 ) {
			pm->ps->movementDir = 4;
		} else if ( pm->cmd.rightmove > 0 && pm->cmd.forwardmove < 0 ) {
			pm->ps->movementDir = 5;
		} else if ( pm->cmd.rightmove > 0 && pm->cmd.forwardmove == 0 ) {
			pm->ps->movementDir = 6;
		} else if ( pm->cmd.rightmove > 0 && pm->cmd.forwardmove > 0 ) {
			pm->ps->movementDir = 7;
		}
	} else {
		// if they aren't actively going directly sideways,
		// change the animation to the diagonal so they
		// don't stop too crooked
		if ( pm->ps->movementDir == 2 ) {
			pm->ps->movementDir = 1;
		} else if ( pm->ps->movementDir == 6 ) {
			pm->ps->movementDir = 7;
		} 
	}
	
	
}


/*
=============
PM_CheckJump
=============
*/
static qboolean PM_CheckJump( void ) {
	if ( pm->ps->pm_flags & PMF_RESPAWNED ) {
		return qfalse;		// don't allow jump until all buttons are up
	}

	if ( pm->cmd.upmove < 10 ) {
		// not holding jump
		return qfalse;
	}
	// must wait for jump to be released
	if ( pm->ps->pm_flags & PMF_JUMP_HELD ) {
		// clear upmove so cmdscale doesn't lower running speed
		pm->cmd.upmove = 0;
		return qfalse;
	}

	pml.groundPlane = qfalse;		// jumping away
	pml.walking = qfalse;
	pm->ps->pm_flags |= PMF_JUMP_HELD;

	pm->ps->groundEntityNum = ENTITYNUM_NONE;
	if(	pm->ps->stats[STAT_JUMPSCALE] >= 0)
		pm->ps->velocity[2] = pm->ps->stats[STAT_JUMPSCALE];
	else
		pm->ps->velocity[2] = JUMP_VELOCITY;
	PM_AddEvent( EV_JUMP );

	if ( pm->cmd.forwardmove >= 0 || pm->cmd.upmove <= 0) {
		PM_ForceLegsAnim( LEGS_JUMP );
		pm->ps->pm_flags &= ~PMF_BACKWARDS_JUMP;
	} else {
		PM_ForceLegsAnim( LEGS_JUMPB );
		pm->ps->pm_flags |= PMF_BACKWARDS_JUMP;
	}

	return qtrue;
}

/*
=============
PM_CheckWaterJump
=============
*/
static qboolean	PM_CheckWaterJump( void ) {
	vec3_t	spot;
	int		cont;
	vec3_t	flatforward;

	if (pm->ps->pm_time) {
		return qfalse;
	}

	// check for water jump
	if ( pm->waterlevel != 2 ) {
		return qfalse;
	}

	flatforward[0] = pml.forward[0];
	flatforward[1] = pml.forward[1];
	flatforward[2] = 0;
	VectorNormalize (flatforward);

	VectorMA (pm->ps->origin, 30, flatforward, spot);
	spot[2] += 4;
	cont = pm->pointcontents (spot, pm->ps->clientNum );
	if ( !(cont & CONTENTS_SOLID) ) {
		return qfalse;
	}

	spot[2] += 16;
	cont = pm->pointcontents (spot, pm->ps->clientNum );
	if ( cont ) {
		return qfalse;
	}

	// jump out of water
	VectorScale (pml.forward, 200, pm->ps->velocity);
	pm->ps->velocity[2] = 350;

	pm->ps->pm_flags |= PMF_TIME_WATERJUMP;
	pm->ps->pm_time = 2000;

	return qtrue;
}

//============================================================================


/*
===================
PM_WaterJumpMove

Flying out of the water
===================
*/
static void PM_WaterJumpMove( void ) {
	// waterjump has no control, but falls

	PM_StepSlideMove( qtrue );

	pm->ps->velocity[2] -= pm->ps->gravity * pml.frametime;
	if (pm->ps->velocity[2] < 0) {
		// cancel as soon as we are falling down again
		pm->ps->pm_flags &= ~PMF_ALL_TIMES;
		pm->ps->pm_time = 0;
	}
}

/*
===================
PM_WaterMove

===================
*/
static void PM_WaterMove( void ) {
	int		i;
	vec3_t	wishvel;
	float	wishspeed;
	vec3_t	wishdir;
	float	scale;
	float	vel;

	if ( PM_CheckWaterJump() ) {
		PM_WaterJumpMove();
		return;
	}
#if 0
	// jump = head for surface
	if ( pm->cmd.upmove >= 10 ) {
		if (pm->ps->velocity[2] > -300) {
			if ( pm->watertype == CONTENTS_WATER ) {
				pm->ps->velocity[2] = 100;
			} else if (pm->watertype == CONTENTS_SLIME) {
				pm->ps->velocity[2] = 80;
			} else {
				pm->ps->velocity[2] = 50;
			}
		}
	}
#endif
	PM_Friction ();

	scale = PM_CmdScale( &pm->cmd );
	//
	// user intentions
	//
	if ( !scale ) {
		wishvel[0] = 0;
		wishvel[1] = 0;
		wishvel[2] = -60;		// sink towards bottom
	} else {
		for (i=0 ; i<3 ; i++)
			wishvel[i] = scale * pml.forward[i]*pm->cmd.forwardmove + scale * pml.right[i]*pm->cmd.rightmove;

		wishvel[2] += scale * pm->cmd.upmove;
	}

	VectorCopy (wishvel, wishdir);
	wishspeed = VectorNormalize(wishdir);

	if ( wishspeed > pm->ps->speed * pm_swimScale ) {
		wishspeed = pm->ps->speed * pm_swimScale;
	}

	PM_Accelerate (wishdir, wishspeed, pm_wateraccelerate);

	// make sure we can go up slopes easily under water
	if ( pml.groundPlane && DotProduct( pm->ps->velocity, pml.groundTrace.plane.normal ) < 0 ) {
		vel = VectorLength(pm->ps->velocity);
		// slide along the ground plane
		PM_ClipVelocity (pm->ps->velocity, pml.groundTrace.plane.normal, 
			pm->ps->velocity, OVERCLIP );

		VectorNormalize(pm->ps->velocity);
		VectorScale(pm->ps->velocity, vel, pm->ps->velocity);
	}

	PM_SlideMove( qfalse );
}



/*
===================
PM_FlyMove

Only with the flight powerup
===================
*/
static void PM_FlyMove( void ) {
	int		i;
	vec3_t	wishvel;
	float	wishspeed;
	vec3_t	wishdir;
	float	scale;

	// normal slowdown
	PM_Friction ();

	scale = PM_CmdScale( &pm->cmd );
	//
	// user intentions
	//
	if ( !scale ) {
		wishvel[0] = 0;
		wishvel[1] = 0;
		wishvel[2] = 0;
	} else {
		for (i=0 ; i<3 ; i++) 
		{
		  wishvel[i] = scale * pml.forward[i]*pm->cmd.forwardmove + scale * pml.right[i]*pm->cmd.rightmove;	
		}
		
		wishvel[2] += scale * pm->cmd.upmove;
		if(pm->cmd.upmove < 0)
		{
		  for (i=0 ; i<3 ; i++) 
		  {
		    wishvel[i] += 0.5 * scale * pml.forward[i]*fabs(DotProduct(pm->ps->velocity, pml.forward));
		  }
		  //wishvel[2] += pm->ps->velocity[2];
		}
	}
	
	VectorCopy (wishvel, wishdir);
	wishspeed = VectorNormalize(wishdir);

	PM_Accelerate (wishdir, wishspeed, pm_flyaccelerate);

	PM_StepSlideMove( qfalse );
}


/*
===================
PM_AirMove

===================
*/
static void PM_AirMove( void ) {
	short		temp = pm->cmd.angles[PITCH];
	int			i;
	vec3_t		wishvel;
	float		fmove, smove;
	vec3_t		wishdir;
	float		wishspeed;
	float		scale;
	usercmd_t	cmd;

	PM_Friction();

	fmove = pm->cmd.forwardmove;
	smove = pm->cmd.rightmove;

	cmd = pm->cmd;
	scale = PM_CmdScale( &cmd );

	// set the movementDir so clients can rotate the legs for strafing
	PM_SetMovementDir();

	// project moves down to flat plane
	pml.forward[2] = 0;
	pml.right[2] = 0;
	VectorNormalize (pml.forward);
	VectorNormalize (pml.right);

	for ( i = 0 ; i < 2 ; i++ ) {
		wishvel[i] = pml.forward[i]*fmove + pml.right[i]*smove;
	}
	wishvel[2] = 0;

	VectorCopy (wishvel, wishdir);
	wishspeed = VectorNormalize(wishdir);
	wishspeed *= scale;

	// not on ground, so little effect on velocity
	PM_Accelerate (wishdir, wishspeed, pm_airaccelerate);
	
	
	if ( temp < 0 && temp > -10000) {
		PM_ContinueLegsAnim( LEGS_JUMPB );
		pm->ps->pm_flags |= PMF_BACKWARDS_JUMP;
	} else if ( temp > 0 && temp < 10000)  {
		PM_ContinueLegsAnim( LEGS_JUMP );
		pm->ps->pm_flags &= ~PMF_BACKWARDS_JUMP;
	}

	// we may have a ground plane that is very steep, even
	// though we don't have a groundentity
	// slide along the steep plane
	if ( pml.groundPlane ) {
		PM_ClipVelocity (pm->ps->velocity, pml.groundTrace.plane.normal, 
			pm->ps->velocity, OVERCLIP );
	}

	PM_StepSlideMove ( qtrue );
}

/*
===================
PM_GrappleMove

===================
*/
static void PM_GrappleMove( void ) {
	vec3_t vel, v;
	float vlen;

	VectorScale(pml.forward, -16, v);
	VectorAdd(pm->ps->grapplePoint, v, v);
	VectorSubtract(v, pm->ps->origin, vel);
	vlen = VectorLength(vel);
	VectorNormalize( vel );

	if (vlen <= 100)
		VectorScale(vel, 10 * vlen, vel);
	else
		VectorScale(vel, 800, vel);

	VectorCopy(vel, pm->ps->velocity);

	pml.groundPlane = qfalse;
}

/*
===================
PM_CrawlMove - this is for Orbb

===================
*/
static void PM_CrawlMove( void ) {
	int			i;
	vec3_t		wishvel;
	float		fmove, smove;
	vec3_t		wishdir;
	float		wishspeed;
	float		scale;
	usercmd_t	cmd;
	float		accelerate;
	float		vel;

	if ( pm->waterlevel > 2 && DotProduct( pml.forward, pml.groundTrace.plane.normal ) > 0 ) {
		// begin swimming
		PM_WaterMove();
		return;
	}


	if ( PM_CheckJump () ) {
		// jumped away
		if ( pm->waterlevel > 1 ) {
			PM_WaterMove();
		} else {
			PM_AirMove();
		}
		return;
	}

	if(!(pm->ps->legsAnim & ANIM_CLIMBWALLS))
		pm->ps->legsAnim |= ANIM_CLIMBWALLS;

	PM_Friction ();

	fmove = pm->cmd.forwardmove;
	smove = pm->cmd.rightmove;

	cmd = pm->cmd;
	scale = PM_CmdScale( &cmd );

	// set the movementDir so clients can rotate the legs for strafing
	PM_SetMovementDir();

	// project the forward and right directions onto the ground plane
	PM_ClipVelocity (pml.forward, pml.groundTrace.plane.normal, pml.forward, OVERCLIP );
	PM_ClipVelocity (pml.right, pml.groundTrace.plane.normal, pml.right, OVERCLIP );
	//
	VectorNormalize (pml.forward);
	VectorNormalize (pml.right);

	for ( i = 0 ; i < 3 ; i++ ) {
		wishvel[i] = pml.forward[i]*fmove + pml.right[i]*smove;
	}
	// when going up or down slopes the wish velocity should Not be zero

	VectorCopy (wishvel, wishdir);
	wishspeed = VectorNormalize(wishdir);
	wishspeed *= scale;

	// clamp the speed lower if ducking
	if ( pm->ps->pm_flags & PMF_DUCKED ) {
		if ( wishspeed > pm->ps->speed * pm_duckScale ) {
			wishspeed = pm->ps->speed * pm_duckScale;
		}
	}

	// clamp the speed lower if wading or walking on the bottom
	if ( pm->waterlevel ) {
		float	waterScale;

		waterScale = pm->waterlevel / 3.0;
		waterScale = 1.0 - ( 1.0 - pm_swimScale ) * waterScale;
		if ( wishspeed > pm->ps->speed * waterScale ) {
			wishspeed = pm->ps->speed * waterScale;
		}
	}

	// when a player gets hit, they temporarily lose
	// full control, which allows them to be moved a bit
	if ( ( pml.groundTrace.surfaceFlags & SURF_SLICK ) || pm->ps->pm_flags & PMF_TIME_KNOCKBACK ) {
		accelerate = pm_airaccelerate;
	} else {
		accelerate = pm_accelerate;
	}

	PM_Accelerate (wishdir, wishspeed, accelerate);

	if ( ( pml.groundTrace.surfaceFlags & SURF_SLICK ) || pm->ps->pm_flags & PMF_TIME_KNOCKBACK ) {
		pm->ps->velocity[2] -= pm->ps->gravity * pml.frametime;
	}

	vel = VectorLength(pm->ps->velocity);

	// slide along the ground plane
	PM_ClipVelocity (pm->ps->velocity, pml.groundTrace.plane.normal, 
		pm->ps->velocity, OVERCLIP );

	// don't decrease velocity when going up or down a slope
	VectorNormalize(pm->ps->velocity);
	VectorScale(pm->ps->velocity, vel, pm->ps->velocity);

	// don't do anything if standing still
	if (DotProduct(pm->ps->velocity, pml.groundTrace.plane.normal) == vel) {
		return;
	}

	PM_SlideMove( qfalse );
			

}

/*
===================
PM_WalkMove

===================
*/
static void PM_WalkMove( void ) {
	int			i;
	vec3_t		wishvel;
	float		fmove, smove;
	vec3_t		wishdir;
	float		wishspeed;
	float		scale;
	usercmd_t	cmd;
	float		accelerate;
	float		vel;

	if ( pm->waterlevel > 2 && DotProduct( pml.forward, pml.groundTrace.plane.normal ) > 0 ) {
		// begin swimming
		PM_WaterMove();
		return;
	}


	if ( PM_CheckJump () ) {
		// jumped away
		if ( pm->waterlevel > 1 ) {
			PM_WaterMove();
		} else {
			PM_AirMove();
		}
		return;
	}

	PM_Friction ();

	fmove = pm->cmd.forwardmove;
	if(fmove < 0)
		fmove *= pm_backScale;
	smove = pm->cmd.rightmove;

	cmd = pm->cmd;
	scale = PM_CmdScale( &cmd );

	// set the movementDir so clients can rotate the legs for strafing
	PM_SetMovementDir();

	// project moves down to flat plane
	pml.forward[2] = 0;
	pml.right[2] = 0;

	// project the forward and right directions onto the ground plane
	PM_ClipVelocity (pml.forward, pml.groundTrace.plane.normal, pml.forward, OVERCLIP );
	PM_ClipVelocity (pml.right, pml.groundTrace.plane.normal, pml.right, OVERCLIP );
	//
	VectorNormalize (pml.forward);
	VectorNormalize (pml.right);

	for ( i = 0 ; i < 3 ; i++ ) {
		wishvel[i] = pml.forward[i]*fmove + pml.right[i]*smove;
	}
	// when going up or down slopes the wish velocity should Not be zero

	VectorCopy (wishvel, wishdir);
	wishspeed = VectorNormalize(wishdir);
	wishspeed *= scale;

	// clamp the speed lower if ducking
	if ( pm->ps->pm_flags & PMF_DUCKED ) {
		if ( wishspeed > pm->ps->speed * pm_duckScale ) {
			wishspeed = pm->ps->speed * pm_duckScale;
		}
	}

	// clamp the speed lower if wading or walking on the bottom
	if ( pm->waterlevel ) {
		float	waterScale;

		waterScale = pm->waterlevel / 3.0;
		waterScale = 1.0 - ( 1.0 - pm_swimScale ) * waterScale;
		if ( wishspeed > pm->ps->speed * waterScale ) {
			wishspeed = pm->ps->speed * waterScale;
		}
	}

	// when a player gets hit, they temporarily lose
	// full control, which allows them to be moved a bit
	if ( ( pml.groundTrace.surfaceFlags & SURF_SLICK ) || pm->ps->pm_flags & PMF_TIME_KNOCKBACK ) {
		accelerate = pm_airaccelerate;
	} else {
		accelerate = pm_accelerate;
	}

	PM_Accelerate (wishdir, wishspeed, accelerate);

	if ( ( pml.groundTrace.surfaceFlags & SURF_SLICK ) || pm->ps->pm_flags & PMF_TIME_KNOCKBACK ) {
		pm->ps->velocity[2] -= pm->ps->gravity * pml.frametime;
	} else {
		// don't reset the z velocity for slopes
	}

	vel = VectorLength(pm->ps->velocity);

	// slide along the ground plane
	PM_ClipVelocity (pm->ps->velocity, pml.groundTrace.plane.normal, 
		pm->ps->velocity, OVERCLIP );

	// don't decrease velocity when going up or down a slope
	VectorNormalize(pm->ps->velocity);
	VectorScale(pm->ps->velocity, vel, pm->ps->velocity);

	// don't do anything if standing still
	if (!pm->ps->velocity[0] && !pm->ps->velocity[1]) {
		return;
	}

	PM_StepSlideMove( qfalse );
	
	//Com_Printf("velocity2 = %1.1f\n", VectorLength(pm->ps->velocity));

}

/*
==============
PM_DeadMove
==============
*/
static void PM_DeadMove( void ) {
	float	forward;

	if ( !pml.walking ) {
		return;
	}

	// extra friction

	forward = VectorLength (pm->ps->velocity);
	forward -= 20;
	if ( forward <= 0 ) {
		VectorClear (pm->ps->velocity);
	} else {
		VectorNormalize (pm->ps->velocity);
		VectorScale (pm->ps->velocity, forward, pm->ps->velocity);
	}
}


/*
===============
PM_NoclipMove
===============
*/
static void PM_NoclipMove( void ) {
	float	speed, drop, friction, control, newspeed;
	int			i;
	vec3_t		wishvel;
	float		fmove, smove;
	vec3_t		wishdir;
	float		wishspeed;
	float		scale;
	short temp = pm->cmd.angles[PITCH];

	pm->ps->viewheight = DEFAULT_VIEWHEIGHT;

	// friction

	speed = VectorLength (pm->ps->velocity);
	if (speed < 1)
	{
		VectorCopy (vec3_origin, pm->ps->velocity);
	}
	else
	{
		drop = 0;

		friction = pm_friction*1.5;	// extra friction
		control = speed < pm_stopspeed ? pm_stopspeed : speed;
		drop += control*friction*pml.frametime;

		// scale the velocity
		newspeed = speed - drop;
		if (newspeed < 0)
			newspeed = 0;
		newspeed /= speed;

		VectorScale (pm->ps->velocity, newspeed, pm->ps->velocity);
	}

	if ( temp < 0 && temp > -16383) {
		PM_ContinueLegsAnim( LEGS_JUMPB );
		pm->ps->pm_flags |= PMF_BACKWARDS_JUMP;
	} else if ( temp > 0 && temp < 16383)  {
		PM_ContinueLegsAnim( LEGS_JUMP );
		pm->ps->pm_flags &= ~PMF_BACKWARDS_JUMP;
	}


	// accelerate
	scale = PM_CmdScale( &pm->cmd );

	fmove = pm->cmd.forwardmove;
	smove = pm->cmd.rightmove;
	
	for (i=0 ; i<3 ; i++)
		wishvel[i] = pml.forward[i]*fmove + pml.right[i]*smove;
	wishvel[2] += pm->cmd.upmove;

	VectorCopy (wishvel, wishdir);
	wishspeed = VectorNormalize(wishdir);
	wishspeed *= scale;

	PM_Accelerate( wishdir, wishspeed, pm_accelerate );

	// move
	VectorMA (pm->ps->origin, pml.frametime, pm->ps->velocity, pm->ps->origin);
}

//============================================================================

/*
================
PM_FootstepForSurface

Returns an event number apropriate for the groundsurface
================
*/
static int PM_FootstepForSurface( void ) {
	if ( pml.groundTrace.surfaceFlags & SURF_NOSTEPS ) {
		return 0;
	}
	if ( pml.groundTrace.surfaceFlags & SURF_METALSTEPS ) {
		return EV_FOOTSTEP_METAL;
	}
	return EV_FOOTSTEP;
}


/*
=================
PM_CrashLand

Check for hard landings that generate sound events
=================
*/
static void PM_CrashLand( void ) {
	float		delta;
	float		dist;
	float		vel, acc;
	float		t;
	float		a, b, c, den;
	float		mass;
	// decide which landing animation to use
	if ( pm->ps->pm_flags & PMF_BACKWARDS_JUMP ) {
		PM_ForceLegsAnim( LEGS_LANDB );
	} else {
		PM_ForceLegsAnim( LEGS_LAND );
	}

	pm->ps->legsTimer = TIMER_LAND;

	// calculate the exact velocity on landing
	dist = pm->ps->origin[2] - pml.previous_origin[2];
	vel = pml.previous_velocity[2];
	acc = -pm->ps->gravity;

	a = acc / 2;
	b = vel;
	c = -dist;

	den =  b * b - 4 * a * c;
	if ( den < 0 ) {
		return;
	}
	t = (-b - sqrt( den ) ) / ( 2 * a );

	delta = vel + t * acc;
	delta = delta*delta * 0.0001;

	// ducking while falling doubles damage
	if ( pm->ps->pm_flags & PMF_DUCKED ) {
		delta *= 2;
	}

	// never take falling damage if completely underwater
	if ( pm->waterlevel == 3 ) {
		return;
	}

	// reduce falling damage if there is standing water
	if ( pm->waterlevel == 2 ) {
		delta *= 0.25;
	}
	if ( pm->waterlevel == 1 ) {
		delta *= 0.5;
	}

	if ( delta < 1 ) {
		return;
	}

	// create a local entity event to play the sound

	// SURF_NODAMAGE is used for bounce pads where you don't ever
	// want to take damage or play a crunch sound
	if(pm->ps->stats[STAT_MASS])
	{
		mass = pm->ps->stats[STAT_MASS];
	}
	else
		mass = 200;
	delta *= sqrt(mass) / 14.0;
	if ( !(pml.groundTrace.surfaceFlags & SURF_NODAMAGE) )  {
		if ( delta > 60 ) {
			BG_AddPredictableEventToPlayerstate( EV_FALL_FAR, delta, pm->ps );
		} else if ( delta > 40 ) {
			// this is a pain grunt, so don't play it if dead
			if ( pm->ps->stats[STAT_HEALTH] > 0 ) 
				BG_AddPredictableEventToPlayerstate( EV_FALL_MEDIUM, delta, pm->ps );
			else
				BG_AddPredictableEventToPlayerstate( EV_FALL_SHORT, delta, pm->ps );
				//PM_AddEvent( EV_FALL_SHORT );
		} else if ( delta > 7 ) {
			PM_AddEvent( EV_FALL_SHORT );
		} else {
			PM_AddEvent( PM_FootstepForSurface() );
		}
	}

	// start footstep cycle over
	pm->ps->bobCycle = 0;
}



/*
=============
PM_CorrectAllSolid
=============
*/
static void PM_CorrectAllSolid( void ) {
	if ( pm->debugLevel ) {
		Com_Printf("%i:allsolid\n", c_pmove);
	}

	// FIXME: jitter around

	pm->ps->groundEntityNum = ENTITYNUM_NONE;
	pml.groundPlane = qfalse;
	pml.walking = qfalse;
}


/*
=============
PM_GroundTraceMissed

The ground trace didn't hit a surface, so we are in freefall
=============
*/
static void PM_GroundTraceMissed( void ) {
	trace_t		trace;
	vec3_t		point;

	if ( pm->ps->groundEntityNum != ENTITYNUM_NONE ) {
		// we just transitioned into freefall
		if ( pm->debugLevel ) {
			Com_Printf("%i:lift\n", c_pmove);
		}

		// if they aren't in a jumping animation and the ground is a ways away, force into it
		// if we didn't do the trace, the player would be backflipping down staircases
		VectorCopy( pm->ps->origin, point );
		point[2] -= 64;

		pm->trace (&trace, pm->ps->origin, pm->mins, pm->maxs, point, pm->ps->clientNum, pm->tracemask);
		if ( trace.fraction == 1.0 ) {
			if ( pm->cmd.forwardmove >= 0 || pm->cmd.upmove <= 0) {
				PM_ForceLegsAnim( LEGS_JUMP );
				pm->ps->pm_flags &= ~PMF_BACKWARDS_JUMP;
			} else {
				PM_ForceLegsAnim( LEGS_JUMPB );
				pm->ps->pm_flags |= PMF_BACKWARDS_JUMP;
			}
		}
	}

	pm->ps->groundEntityNum = ENTITYNUM_NONE;
	pml.groundPlane = qfalse;
	pml.walking = qfalse;
}

extern vec3_t	bytedirs[NUMVERTEXNORMALS];
/*
=============
PM_GroundWallTrace
=============
*/
static void PM_GroundWallTrace( void ) {
	vec3_t		point, forward, right, up;
	vec3_t mins, maxs;
	trace_t		trace;
	int i;

	
	AngleVectors(pm->ps->viewangles, forward, right, up);
	mins[0] = -15;
	mins[1] = -15;
	mins[2] = -15;
	maxs[0] = 15;
	maxs[1] = 15;
	maxs[2] = 15;
	pml.groundTrace.fraction = 1.1;
	for(i = 0; i < 5; i++)
	{
		VectorCopy(pm->ps->origin, point);
			
		switch (i)
		{
		case 0:
			point[2] += 0.25;
			break;
		case 1:
			point[1] += 0.25;
			break;
		case 2:
			point[1] -= 0.25;
			break;
		case 3:
			point[0] += 0.25;
			break;
		case 4:
			point[0] -= 0.25;
			break;
		case 5:
			point[2] -= 0.25;
			break;
		}
		pm->trace (&trace, pm->ps->origin, pm->mins, pm->maxs, point, pm->ps->clientNum, pm->tracemask);
			
		if(trace.fraction < 1.0 && !(trace.surfaceFlags & (SURF_SKY | SURF_NOIMPACT)))
		{
			pml.groundTrace = trace;
			break;
		}
		else if ( trace.allsolid ) 
		{// do something corrective if the trace starts in a solid...
			PM_CorrectAllSolid();
			return;
		}
	}
	

	// if the trace didn't hit anything, we are in free fall
	if(pml.groundTrace.fraction >= 1.0 || (trace.surfaceFlags & (SURF_SKY | SURF_NOIMPACT)))
	{		
		pml.groundPlane = qfalse;
		pml.walking = qfalse;
		return;
//		PM_GroundTraceMissed();
	}
	
	
	pml.groundPlane = qtrue;
	pml.walking = qtrue;
	if(!(pm->ps->legsAnim & ANIM_CLIMBWALLS))
		pm->ps->legsAnim |= ANIM_CLIMBWALLS;

	// hitting solid ground will end a waterjump
	if (pm->ps->pm_flags & PMF_TIME_WATERJUMP)
	{
		pm->ps->pm_flags &= ~(PMF_TIME_WATERJUMP | PMF_TIME_LAND);
		pm->ps->pm_time = 0;
	}

	pm->ps->groundEntityNum = trace.entityNum;
	pml.groundTrace = trace;
	PM_AddTouchEnt( trace.entityNum );
}

static qboolean PM_LadderTrace( void ) {
	vec3_t		point, forward, right, up;
	vec3_t mins, maxs;
	trace_t		trace;
	int i;

	
	AngleVectors(pm->ps->viewangles, forward, right, up);
	mins[0] = -15;
	mins[1] = -15;
	mins[2] = -15;
	maxs[0] = 15;
	maxs[1] = 15;
	maxs[2] = 15;
	pml.groundTrace.fraction = 1.1;
	for(i = 0; i < 5; i++)
	{
		VectorCopy(pm->ps->origin, point);
			
		switch (i)
		{
		case 0:
			point[2] += 0.25;
			break;
		case 1:
			point[1] += 0.25;
			break;
		case 2:
			point[1] -= 0.25;
			break;
		case 3:
			point[0] += 0.25;
			break;
		case 4:
			point[0] -= 0.25;
			break;
		case 5:
			point[2] -= 0.25;
			break;
		}
		pm->trace (&trace, pm->ps->origin, pm->mins, pm->maxs, point, pm->ps->clientNum, pm->tracemask);
			
		if(trace.fraction < 1.0 && (trace.surfaceFlags & SURF_LADDER))
		{
			pml.groundTrace = trace;
			break;
		}
		else if ( trace.allsolid ) 
		{// do something corrective if the trace starts in a solid...
			PM_CorrectAllSolid();
			return qfalse;
		}
	}
	

	// if the trace didn't hit anything, we are in free fall
	if(pml.groundTrace.fraction >= 1.0 || (!trace.surfaceFlags & SURF_LADDER))
		return qfalse;
	
	
	pml.groundPlane = qtrue;
	pml.walking = qtrue;
	if(!(pm->ps->legsAnim & ANIM_CLIMBWALLS))
		pm->ps->legsAnim |= ANIM_CLIMBWALLS;

	// hitting solid ground will end a waterjump
	if (pm->ps->pm_flags & PMF_TIME_WATERJUMP)
	{
		pm->ps->pm_flags &= ~(PMF_TIME_WATERJUMP | PMF_TIME_LAND);
		pm->ps->pm_time = 0;
	}

	pm->ps->groundEntityNum = trace.entityNum;
	pml.groundTrace = trace;
	PM_AddTouchEnt( trace.entityNum );
	return qtrue;
}

/*
=============
PM_GroundTrace
=============
*/
static void PM_GroundTrace( void ) {
	vec3_t		point;
	trace_t		trace;
	
	if((pm->ps->stats[STAT_EXTENDED_INFO] & EXT_WALLCLIMBER) && (pm->cmd.upmove < 0))
	{		
		if(!(pm->ps->legsAnim & ANIM_CLIMBWALLS))
			pm->ps->legsAnim |= ANIM_CLIMBWALLS;
		PM_GroundWallTrace();			
		if(pml.groundPlane)
			return;
	}
			
	point[0] = pm->ps->origin[0];
	point[1] = pm->ps->origin[1];
	point[2] = pm->ps->origin[2] - 0.25;

	pm->trace (&trace, pm->ps->origin, pm->mins, pm->maxs, point, pm->ps->clientNum, pm->tracemask);
	pml.groundTrace = trace;

	// do something corrective if the trace starts in a solid...
	if ( trace.allsolid ) {
		PM_CorrectAllSolid();
		return;
	}

	// if the trace didn't hit anything, we are in free fall
	if ( trace.fraction == 1.0 && !PM_LadderTrace()) {		
		PM_GroundTraceMissed();
		pml.groundPlane = qfalse;
		pml.walking = qfalse;		
		return;
	}

	// check if getting thrown off the ground
	if(pml.groundTrace.surfaceFlags & SURF_LADDER)
	{
	}
	else if ( pm->ps->velocity[2] > 0 && DotProduct( pm->ps->velocity, trace.plane.normal ) > 10 ) {
		if ( pm->debugLevel ) {
			Com_Printf("%i:kickoff\n", c_pmove);
		}
		// go into jump animation
		if ( pm->cmd.forwardmove >= 0 || pm->cmd.upmove <= 0) {
			PM_ForceLegsAnim( LEGS_JUMP );
			pm->ps->pm_flags &= ~PMF_BACKWARDS_JUMP;
		} else {
			PM_ForceLegsAnim( LEGS_JUMPB );
			pm->ps->pm_flags |= PMF_BACKWARDS_JUMP;
		}

		pm->ps->groundEntityNum = ENTITYNUM_NONE;
		pml.groundPlane = qfalse;
		pml.walking = qfalse;
		return;
	}
	
	// slopes that are too steep will not be considered onground
	else if ( trace.plane.normal[2] < MIN_WALK_NORMAL ) {
		if ( pm->debugLevel ) {
			Com_Printf("%i:steep\n", c_pmove);
		}
		// FIXME: if they can't slide down the slope, let them
		// walk (sharp crevices)
		if(pm->ps->groundEntityNum == ENTITYNUM_NONE)
		{
			PM_CrashLand();
		}
		else
			pm->ps->groundEntityNum = ENTITYNUM_NONE;
		pml.groundPlane = qtrue;
		pml.walking = qfalse;
		return;
	}

	pml.groundPlane = qtrue;
	pml.walking = qtrue;

	// hitting solid ground will end a waterjump
	if (pm->ps->pm_flags & PMF_TIME_WATERJUMP)
	{
		pm->ps->pm_flags &= ~(PMF_TIME_WATERJUMP | PMF_TIME_LAND);
		pm->ps->pm_time = 0;
	}

	if ( pm->ps->groundEntityNum == ENTITYNUM_NONE ) {
		// just hit the ground
		if ( pm->debugLevel ) {
			Com_Printf("%i:Land\n", c_pmove);
		}
		
		PM_CrashLand();

		// don't do landing time if we were just going down a slope
		if ( pml.previous_velocity[2] < -50 ) {
			// don't allow another jump for a little while
			pm->ps->pm_flags |= PMF_TIME_LAND;
			pm->ps->pm_time = pm->ps->stats[STAT_JUMP_PAUSE];
			if(pm->ps->pm_time < 10)
				pm->ps->pm_time = 10;
		}
	}

	pm->ps->groundEntityNum = trace.entityNum;

	// don't reset the z velocity for slopes
//	pm->ps->velocity[2] = 0;

	PM_AddTouchEnt( trace.entityNum );
}


/*
=============
PM_SetWaterLevel	FIXME: avoid this twice?  certainly if not moving
=============
*/
static void PM_SetWaterLevel( void ) {
	vec3_t		point;
	int			cont;
	int			sample1;
	int			sample2;

	//
	// get waterlevel, accounting for ducking
	//
	pm->waterlevel = 0;
	pm->watertype = 0;

	point[0] = pm->ps->origin[0];
	point[1] = pm->ps->origin[1];
	point[2] = pm->ps->origin[2] + MINS_Z + 1;	
	cont = pm->pointcontents( point, pm->ps->clientNum );

	if ( cont & MASK_WATER ) {
		sample2 = pm->ps->viewheight - MINS_Z;
		sample1 = sample2 / 2;

		pm->watertype = cont;
		pm->waterlevel = 1;
		point[2] = pm->ps->origin[2] + MINS_Z + sample1;
		cont = pm->pointcontents (point, pm->ps->clientNum );
		if ( cont & MASK_WATER ) {
			pm->waterlevel = 2;
			point[2] = pm->ps->origin[2] + MINS_Z + sample2;
			cont = pm->pointcontents (point, pm->ps->clientNum );
			if ( cont & MASK_WATER ){
				pm->waterlevel = 3;
			}
		}
	}

}



/*
==============
PM_CheckDuck

Sets mins, maxs, and pm->ps->viewheight
==============
*/
static void PM_CheckDuck (void)
{
	trace_t	trace;
	

	pm->mins[0] = -15;
	pm->mins[1] = -15;

	pm->maxs[0] = 15;
	pm->maxs[1] = 15;

	pm->mins[2] = MINS_Z;

	if (pm->ps->pm_type == PM_DEAD)
	{
		pm->maxs[2] = -8;
		pm->ps->viewheight = DEAD_VIEWHEIGHT;
		return;
	}
	
	if (pm->cmd.upmove < 0)
	{	// duck
		pm->ps->pm_flags |= PMF_DUCKED;
	}
	else
	{	// stand up if possible
		if (pm->ps->pm_flags & PMF_DUCKED)
		{
			// try to stand up
			pm->maxs[2] = 32;
			pm->trace (&trace, pm->ps->origin, pm->mins, pm->maxs, pm->ps->origin, pm->ps->clientNum, pm->tracemask );
			if (!trace.allsolid)
				pm->ps->pm_flags &= ~PMF_DUCKED;
		}
	}

	if(pm->ps->stats[STAT_EXTENDED_INFO] & EXT_WALLCLIMBER)
	{
		pm->maxs[2] = 24;
		pm->ps->viewheight = 4;
	}
	else if (pm->ps->pm_flags & PMF_DUCKED)
	{
		pm->maxs[2] = 16;
		pm->ps->viewheight = CROUCH_VIEWHEIGHT;
	}
	else
	{
		pm->maxs[2] = 32;
		pm->ps->viewheight = DEFAULT_VIEWHEIGHT;
	}
}



//===================================================================


/*
===============
PM_Footsteps
===============
*/
static void PM_Footsteps( void ) {
	float		bobmove;
	int			old;
	qboolean	footstep;
	//
	// calculate speed and cycle to be used for
	// all cyclic walking effects
	//

	if(pm->ps->powerups[PW_FLIGHT])
	{
		pm->xyspeed = sqrt( pm->ps->velocity[0] * pm->ps->velocity[0] +  pm->ps->velocity[1] * pm->ps->velocity[1] + pm->ps->velocity[2] * pm->ps->velocity[2]);
		PM_ContinueLegsAnim( LEGS_SWIM );
	}
	else if(((pm->ps->stats[STAT_EXTENDED_INFO] & EXT_WALLCLIMBER) || pml.groundTrace.surfaceFlags & SURF_LADDER) && ( pml.groundPlane ) )
	{
		pm->xyspeed = DotProduct(pm->ps->velocity, pml.groundTrace.plane.normal);//sqrt( pm->ps->velocity[0] * pm->ps->velocity[0] +  pm->ps->velocity[1] * pm->ps->velocity[1] );
		if((pm->ps->stats[STAT_EXTENDED_INFO] & EXT_WALLCLIMBER) && !(pm->ps->legsAnim & ANIM_CLIMBWALLS))
			pm->ps->legsAnim |= ANIM_CLIMBWALLS;

	}
	else
		pm->xyspeed = sqrt( pm->ps->velocity[0] * pm->ps->velocity[0] +  pm->ps->velocity[1] * pm->ps->velocity[1] );
	

	if ( pm->ps->groundEntityNum == ENTITYNUM_NONE ) {
		// airborne leaves position in cycle intact, but doesn't advance
		if ( pm->waterlevel > 1 ) {
			PM_ContinueLegsAnim( LEGS_SWIM );
		}
		return;
	}

	// if not trying to move
	if ( !pm->cmd.forwardmove && !pm->cmd.rightmove ) {
		if (  pm->xyspeed < 5 ) {
			pm->ps->bobCycle = 0;	// start at beginning of cycle again
			if ( pm->ps->pm_flags & PMF_DUCKED ) 
			{
				PM_ContinueLegsAnim( LEGS_IDLECR );
			} else {
				PM_ContinueLegsAnim( LEGS_IDLE );
			}
			if(pm->ps->stats[STAT_EXTENDED_INFO] & EXT_WALLCLIMBER)
				if(!(pm->ps->legsAnim & ANIM_CLIMBWALLS))
					pm->ps->legsAnim |= ANIM_CLIMBWALLS;			
		}
		return;
	}
	

	footstep = qfalse;
	if ( pm->ps->pm_flags & PMF_DUCKED ) {
		bobmove = 0.5;	// ducked characters bob much faster
		PM_ContinueLegsAnim( LEGS_WALKCR );
		// ducked characters never play footsteps
	} else 	if ( pm->ps->pm_flags & PMF_BACKWARDS_RUN ) 
	{
		bobmove = pm->xyspeed / 640;
		if ( !( pm->cmd.buttons & BUTTON_WALKING ) ) 
			footstep = qtrue;
		PM_ContinueLegsAnim( LEGS_BACK );
	} else {
		if ( !(pm->cmd.buttons & BUTTON_WALKING ) ) {
			bobmove = pm->xyspeed / 500;//0.4;	// faster speeds bob faster
			PM_ContinueLegsAnim( LEGS_RUN );
			footstep = qtrue;
		} else {
			bobmove = pm->xyspeed / 640;	// walking bobs slow
			PM_ContinueLegsAnim( LEGS_WALK );
		}
	}

	if(pm->ps->stats[STAT_EXTENDED_INFO] & EXT_WALLCLIMBER)
	{
		if(!(pm->ps->legsAnim & ANIM_CLIMBWALLS))
			pm->ps->legsAnim |= ANIM_CLIMBWALLS;
	}
	// check for footstep / splash sounds
	old = pm->ps->bobCycle;
	pm->ps->bobCycle = (int)( old + bobmove * pml.msec ) & 255;

	// if we just crossed a cycle boundary, play an apropriate footstep event
	if ( ( ( old + 64 ) ^ ( pm->ps->bobCycle + 64 ) ) & 128 ) {
		if ( pm->waterlevel == 0 ) {
			// on ground will only play sounds if running
			if ( footstep && !pm->noFootsteps ) {
				PM_AddEvent( PM_FootstepForSurface() );
			}
		} else if ( pm->waterlevel == 1 ) {
			// splashing
			PM_AddEvent( EV_FOOTSPLASH );
		} else if ( pm->waterlevel == 2 ) {
			// wading / swimming at surface
			PM_AddEvent( EV_SWIM );
		} else if ( pm->waterlevel == 3 ) {
			// no sound when completely underwater

		}
	}
}

/*
==============
PM_WaterEvents

Generate sound events for entering and leaving water
==============
*/
static void PM_WaterEvents( void ) {		// FIXME?
	//
	// if just entered a water volume, play a sound
	//
	if (!pml.previous_waterlevel && pm->waterlevel) {
		PM_AddEvent( EV_WATER_TOUCH );
	}

	//
	// if just completely exited a water volume, play a sound
	//
	if (pml.previous_waterlevel && !pm->waterlevel) {
		PM_AddEvent( EV_WATER_LEAVE );
	}

	//
	// check for head just going under water
	//
	if (pml.previous_waterlevel != 3 && pm->waterlevel == 3) {
		PM_AddEvent( EV_WATER_UNDER );
	}

	//
	// check for head just coming out of water
	//
	if (pml.previous_waterlevel == 3 && pm->waterlevel != 3) {
		PM_AddEvent( EV_WATER_CLEAR );
	}
}


/*
===============
PM_BeginWeaponChange
===============
*/

static void PM_BeginWeaponChange(  ) 
{
	
	if ( pm->ps->weaponstate == WEAPON_DROPPING ) 
	{
		return;
	}

	pm->ps->weaponstate = WEAPON_DROPPING;
	pm->ps->weaponTime += 200;
	PM_StartTorsoAnim( TORSO_DROP );
}


/*
===============
PM_FinishWeaponChange
===============
*/
static void PM_LoadAmmoIntoWeapon(int *weapon, int *wammo, int *ammo, int *iqty)
{ // ahh, pointers. I can never forget order of operation on them, so I'm just using wammo[0] instead of *wammo, just to be sure.
	gitem_t *witem, *aitem;

	witem = bg_itemlist + *weapon;
	aitem = bg_itemlist + *ammo;

	if( witem->giTag == WP_ROCKET_LAUNCHER || witem->giUses == AMMO_BULLETS)
	{ // rocket launcher and smg/chaingun belts can be as long as we want.
		PM_AddEvent(EV_RELOAD);
		wammo[0] += iqty[0];
		iqty[0] = 0;
		ammo[0] = 0;
		
	}
	else if((witem->giTag == WP_GRENADE_LAUNCHER || witem->giUses == AMMO_SHELLS ) && wammo[0] < witem->quantity)
	{ // one at a time, please.
		PM_AddEvent(EV_RELOAD);
		wammo[0]++;
		iqty[0]--;
		if(witem->giTag == WP_SHOTGUN && iqty[0])
		{ // well, two for the double barreled shotgun
			wammo[0]++;
			iqty[0]--;
			}
		if(!iqty[0])
			ammo[0] = 0;						
		
	}
	else if(witem->giUses == AMMO_CELLS || witem->giUses == AMMO_NAPALM || witem->giUses == AMMO_SLUGS )
	{ // energy and flame weapons can only be reloaded when empty.
		int old = wammo[0];
		PM_AddEvent(EV_RELOAD);
		if(old == -1)
		{ // we didn't even have a clip in the gun
			old = 0;
			ammo[0] = 0;
		}
		else // put the clip in the gun back on our belt
			ammo[0] = BG_FindAmmoForWeapon(witem->giTag) - bg_itemlist;
		switch(witem->giTag)
		{
		default: // 1 energy per shot
			wammo[0] = iqty[0];
			iqty[0] = old;
			break;
		case WP_LIGHTNING: // 2 energy per shot
			wammo[0] = iqty[0] * 2;
			iqty[0] = old / 2;
			break;
		case WP_GAUNTLET: // 4 energy per shot			
			wammo[0] = iqty[0] * 4;
			iqty[0] = old / 4;
			break;
		}
		
	}
}

static void PM_FinishWeaponChange( void ) {
	gitem_t *inventory, *inhand;
	int		selected, iqty, hqty;//weapon = WP_NONE, i, other = 0, ammo, qty, selected;

	
	// we are now at the state where we need to decide what we're doing.
	// NOTE: This is the meat of our inventory system. It's in pmove because it should be predictable.
	// Magic happens here. If you don't understand what's going on, DON'T MESS WITH IT.

	// note that all inventory slots are indexes into bg_itemlist; this is so you can hold anything. -- FIXME: Make indexes into g_entities?

	selected = pm->cmd.weapon * 2; // every other ammo slot is now an inventory slot; the next slot up is its ammo/quantity/whatever
	inhand = &bg_itemlist[pm->ps->weapon]; // pm->ps->weapon is also an inventory slot (it's what you're holding right now)
	hqty = pm->ps->ammo[AMMO_CURRENT];//pm->ps->ammo[0] is ammo/quantity/whatever for pm->ps->weapon -- FIXME: make 0 mirror weapon and 1 be ammo for consistency?
	inventory = &bg_itemlist[pm->ps->ammo[selected]]; // ammo[2,4,6,8,10,12,14] are all indexes into bg_itemlist
	iqty = pm->ps->ammo[selected + 1];
	
	if(selected == 0 || selected > AMMO_ITEM_BIG) // no inventory slot selected
	{ // item selected is what's in our hands - drop it
		BG_AddPredictableEventToPlayerstate( EV_DROP_WEAPON, pm->ps->weapon, pm->ps );	
	}
	else if(selected == AMMO_ITEM_BIG) // big inventory slot [7] selected
	{ // on our back - this is a special case, because only weapons should go here. 
		
		if(!inventory->giType && !inhand->giType && pm->ps->stats[STAT_ARMOR]) // nothing in the inventory, but we're wearing armor
		{ // we want to take our armor off
			pm->ps->weapon = BG_FindItem("Armor") - bg_itemlist;
			pm->ps->ammo[AMMO_CURRENT] = pm->ps->stats[STAT_ARMOR];

			pm->ps->stats[STAT_ARMOR] = 0;
		}
		else if(!inhand->giType || inhand->giType == IT_WEAPON || (inhand->giType == IT_ARMOR && !inventory->giType && pm->ps->stats[STAT_ARMOR] == 0))
		{ // nothing in our hands, or we want to put armor on - either way, swap what's in our hands with whats in inventory
			if(inhand->giType == IT_ARMOR)
			{ // if it's armor, it goes on our body instead of on our back
				pm->ps->stats[STAT_ARMOR] = hqty;
				hqty = 0;
				inhand = bg_itemlist;
			}
			pm->ps->ammo[selected] = inhand - bg_itemlist;
			pm->ps->ammo[selected + 1] = hqty;
			pm->ps->ammo[AMMO_CURRENT] = iqty;
			pm->ps->weapon = inventory - bg_itemlist;
			if(pm->ps->weapon)
				BG_AddPredictableEventToPlayerstate( EV_ITEM_PICKUP, pm->ps->weapon, pm->ps);
		}
		else if(inhand->giType == IT_AMMO && inventory->giType == IT_WEAPON && inhand->giTag == inventory->giUses)
		{ // load this ammo into the gun on our back (kinda tricky - uses pointers)
			PM_LoadAmmoIntoWeapon(&pm->ps->ammo[selected], &pm->ps->ammo[selected + 1], &pm->ps->weapon, &pm->ps->ammo[AMMO_CURRENT]);	
		}
	}
	else if(inhand->giType == IT_HEALTH)
	{ // health can always fit into a slot
		pm->ps->ammo[selected] = inhand - bg_itemlist;
		pm->ps->ammo[selected + 1] = hqty;
		pm->ps->ammo[AMMO_CURRENT] = iqty;
		pm->ps->weapon = inventory - bg_itemlist;
	}
	else if(inhand->giType == IT_AMMO)
	{	// we're holding ammo	
		if(inhand->giTag == AMMO_BULLETS || inhand->giTag == AMMO_ROCKETS )
		{ // bullets and rockets are belt-fed
			if(inventory->giType == IT_AMMO && inventory->giTag == inhand->giTag)//inventory matches what's in our hands
			{ //put it back on the belt in our hands
				pm->ps->ammo[selected] = 0;
				pm->ps->ammo[selected + 1] = 0;
				pm->ps->ammo[AMMO_CURRENT] += iqty;
			}
			else 
			{// swap with what's on our belt
				int max;
				if(inhand->giTag == AMMO_BULLETS)
					max = 20;// can only fit 20 bullets in a slot
				else
					max = 1;// can only fit 1 rocket in a slot
				if(hqty <= max)
				{ // swap with what's on our belt if it'll fit
					pm->ps->ammo[selected] = inhand - bg_itemlist;
					pm->ps->ammo[selected + 1] = hqty;
					pm->ps->ammo[AMMO_CURRENT] = iqty;
					pm->ps->weapon = inventory - bg_itemlist;
				}
				else if(!inventory->giType)
				{ // stick as much as we can on our belt
					pm->ps->ammo[selected] = inhand - bg_itemlist;
					pm->ps->ammo[selected + 1] = max;
					pm->ps->ammo[AMMO_CURRENT] -= max;
				}
			}
		}		
		else // isn't a belt; must be a clip or just a jumble of shells
		{ // swap with what's on our belt
			pm->ps->ammo[selected] = inhand - bg_itemlist;
			pm->ps->ammo[selected + 1] = hqty;
			pm->ps->ammo[AMMO_CURRENT] = iqty;
			pm->ps->weapon = inventory - bg_itemlist;
		}
	}
	else if(inhand->giType == inventory->giType)
	{ // we can pretty much guarantee they're compatible
		pm->ps->ammo[selected] = inhand - bg_itemlist;
		pm->ps->ammo[selected + 1] = hqty;
		pm->ps->ammo[AMMO_CURRENT] = iqty;
		pm->ps->weapon = inventory - bg_itemlist;
		if(pm->ps->weapon)
			BG_AddPredictableEventToPlayerstate( EV_ITEM_PICKUP, pm->ps->weapon, pm->ps);
	}
	else if(!inhand->giType)
	{ // hands are empty - put whatever's in inventory into our hands
		pm->ps->weapon = pm->ps->ammo[selected];
		pm->ps->ammo[AMMO_CURRENT] = iqty;
		pm->ps->ammo[selected] = 0;
		pm->ps->ammo[selected + 1] = 0;
		if(pm->ps->weapon)
			BG_AddPredictableEventToPlayerstate( EV_ITEM_PICKUP, pm->ps->weapon, pm->ps);
	}
	else if(!inventory->giType)
	{ // empty slot - put whatever's in our hands into the slot, after checking to see if it will fit.
		int empty;
		if(inhand->giType == IT_WEAPON) // we can't put a weapon into a belt slot, but we can fit its ammo
		{	// some weapons are empty at 0 (belt-fed, etc.); 
			// others are empty at -1 (clip-fed, where 0 means the clip is empty, but -1 means there isn't a clip at all)
			switch(inhand->giUses)
			{ // clip-fed
			case AMMO_CELLS:
			case AMMO_NAPALM:
			case AMMO_SLUGS:
				empty = -1;
				break;
			default: // belt-fed or internal magazine
				empty = 0;
				break;
			}
		}
		if(inhand->giType == IT_WEAPON && hqty == empty)
		{ // no ammo (not even an empty clip) to put in the slot
			PM_AddEvent(EV_NOAMMO);			
		}
		else if(inhand->giType == IT_WEAPON || (inhand->giType == IT_AMMO))
		{ // both use basically the same code, since in both cases you end up with ammo in your pockets
			int ammo, itemx;
			if(inhand->giType == IT_AMMO )
			{
				ammo = inhand->giTag;
				itemx = inhand - bg_itemlist;
			}
			else
			{// can't fit the whole gun into a pocket, so let's unload the ammo.			
				ammo = inhand->giUses;
				itemx = BG_FindAmmoForWeapon(inhand->giTag) - bg_itemlist;
				if(ammo == AMMO_CELLS)
				{
					if(inhand->giTag == WP_LIGHTNING)
						hqty /= 2;
					else if(inhand->giTag == WP_GAUNTLET)
						hqty /= 4;
				}
			}

			PM_AddEvent(EV_RELOAD);
			switch(ammo)
			{
			case AMMO_SHELLS: // 1 qty = 1 shot
				pm->ps->ammo[AMMO_CURRENT] = 0;
				pm->ps->ammo[selected] = itemx;
				pm->ps->ammo[selected + 1] = hqty;
				break;
			case AMMO_SLUGS:
			case AMMO_NAPALM:
			case AMMO_CELLS:
				pm->ps->ammo[AMMO_CURRENT] = -1;
				pm->ps->ammo[selected] = itemx;
				pm->ps->ammo[selected + 1] = hqty;
				break;
				
			case AMMO_GRENADES: // only one will fit per slot
			case AMMO_ROCKETS: 
				pm->ps->ammo[AMMO_CURRENT]--;
				pm->ps->ammo[selected] = itemx;
				pm->ps->ammo[selected + 1] = 1;
				break;
			case AMMO_BULLETS: // can fit up to 20
				if(hqty > 20)
					hqty = 20;
				pm->ps->ammo[AMMO_CURRENT] -= hqty;
				pm->ps->ammo[selected] = itemx;
				pm->ps->ammo[selected + 1] = hqty;
				break;
			default:
				break; // we can't fit anything else in a pocket
			}
			if(inhand->giType == IT_AMMO && pm->ps->ammo[AMMO_CURRENT] <= 0)
			{ // we ran out
				pm->ps->weapon = 0;
			}
			
		}
		else if(inhand->giType != IT_WEAPON && inhand->giType != IT_ARMOR)
		{ // non-weapons go in any other slot
			BG_AddPredictableEventToPlayerstate( EV_ITEM_PICKUP, pm->ps->weapon, pm->ps);
			pm->ps->ammo[selected] = pm->ps->weapon;
			pm->ps->ammo[selected + 1] = hqty;
			pm->ps->ammo[AMMO_CURRENT] = 0;
			pm->ps->weapon = 0;						
		}

	}
	else if(inhand->giType == IT_WEAPON && inventory->giType == IT_AMMO && inhand->giUses == inventory->giTag)
	{ // item selected is ammo for current weapon - load it into the gun
		BG_AddPredictableEventToPlayerstate( EV_ITEM_PICKUP, pm->ps->weapon, pm->ps);
		PM_LoadAmmoIntoWeapon(&pm->ps->weapon, &pm->ps->ammo[AMMO_CURRENT], &pm->ps->ammo[selected], &pm->ps->ammo[selected + 1]);
	}
	pm->ps->weaponstate = WEAPON_RAISING;
	pm->ps->weaponTime += 250;
	PM_StartTorsoAnim( TORSO_RAISE );
	
	
}

/*
==============
PM_TorsoAnimation

==============
*/
static void PM_TorsoAnimation( void ) {
	if ( pm->ps->weaponstate == WEAPON_READY ) {
		if (bg_itemlist[pm->ps->weapon].giType == IT_WEAPON && (bg_itemlist[pm->ps->weapon].giTag > WP_GAUNTLET && bg_itemlist[pm->ps->weapon].giTag < WP_REDFLAG))
			PM_ContinueTorsoAnim( TORSO_STAND );
		else if(bg_itemlist[pm->ps->weapon].giType == IT_AMMO && bg_itemlist[pm->ps->weapon].giTag == AMMO_GRENADES && (pm->cmd.buttons & BUTTON_ATTACK))
			PM_ContinueTorsoAnim( TORSO_DROP );
		else
			PM_ContinueTorsoAnim( TORSO_STAND2 );

		return;
	}
}


/*
==============
PM_Weapon

Generates weapon events and modifes the weapon counter
==============
*/
static void PM_Weapon( void ) {
	int		addTime;

	// don't allow attack until all buttons are up
	if ( pm->ps->pm_flags & PMF_RESPAWNED ) 
	{
		return;
	}

	// ignore if spectator
	if ( pm->ps->persistant[PERS_TEAM] == TEAM_SPECTATOR ) 
	{
		return;
	}

	// check for dead player
	if ( pm->ps->stats[STAT_HEALTH] <= 0 ) 
	{
		//pm->ps->weapon = 0;
		return;
	}

	// check for item using
	if ( pm->cmd.buttons & BUTTON_USE_HOLDABLE ) 
	{
		if ( ! ( pm->ps->pm_flags & PMF_USE_ITEM_HELD ) ) 
		{
			if(pm->ps->weaponstate != WEAPON_READY || (pm->ps->weapon == 0 && pm->cmd.weapon == 0))
			{
				pm->ps->pm_flags |= PMF_USE_ITEM_HELD;
			}
			else
			{ // ## Hentai ## - Real simple, BG_FinishWeaponChange will handle the gritty details.
				PM_BeginWeaponChange();
			}
			return;
		}
	} 
	else if(pm->ps->weaponTime <= 0)
	{
		pm->ps->pm_flags &= ~PMF_USE_ITEM_HELD;
	}

	// make weapon function
	if ( pm->ps->weaponTime > 0 ) {
		pm->ps->weaponTime -= pml.msec;
	}

	// check for weapon change
	// can't change if weapon is firing, but can change
	// again if lowering or raising
	
	//if ((pm->ps->weaponstate == WEAPON_READY) && ( pm->cmd.weapon ) )
	//	PM_BeginWeaponChange();

	
	if ( pm->ps->weaponTime > 0) {
		return;
	}
	
	// change weapon if time
	if ( pm->ps->weaponstate == WEAPON_DROPPING ) {
		
		PM_FinishWeaponChange();
		return;
	}

	if ( pm->ps->weaponstate == WEAPON_RAISING ) {
		pm->ps->weaponstate = WEAPON_READY;
		return;
	}

	if(bg_itemlist[pm->ps->weapon].giType == IT_AMMO && bg_itemlist[pm->ps->weapon].giTag == AMMO_GRENADES && pm->ps->ammo[AMMO_CURRENT] == 1)
	{
		if(pm->ps->pm_flags & PMF_FIRE_HELD)
		{
			if(pm->cmd.buttons & BUTTON_ATTACK)
			{
				pm->ps->weaponTime = 0;
				pm->ps->weaponstate = WEAPON_READY;
			}
			else
			{
				pm->ps->pm_flags &= ~PMF_FIRE_HELD;
				pm->ps->weaponstate = WEAPON_FIRING;
				pm->ps->weaponTime = 600;
				PM_AddEvent( EV_FIRE_WEAPON );
				PM_StartTorsoAnim( TORSO_ATTACK2 );
			}
		}
		else if(pm->cmd.buttons & BUTTON_ATTACK)
		{
			pm->ps->pm_flags |= PMF_FIRE_HELD;
			PM_AddEvent( EV_NOAMMO ); // just to make the 'chink!' sound
			PM_StartTorsoAnim( TORSO_DROP );
		}
		return;
	}
	
	// check for fire
	
	
	if ( ! (pm->cmd.buttons & BUTTON_ATTACK) ) {
		pm->ps->weaponTime = 0;
		pm->ps->weaponstate = WEAPON_READY;
		pm->ps->pm_flags &= ~PMF_FIRE_HELD;
		return;
	}
	pm->ps->pm_flags |= PMF_FIRE_HELD;

	// start the animation even if out of ammo
	if ( (bg_itemlist[pm->ps->weapon].giType != IT_WEAPON && bg_itemlist[pm->ps->weapon].giType != IT_HEALTH) || (bg_itemlist[pm->ps->weapon].giType == IT_WEAPON && (bg_itemlist[pm->ps->weapon].giTag <= WP_GAUNTLET || bg_itemlist[pm->ps->weapon].giTag >= WP_REDFLAG))) {
		// the guantlet and fist only "fire" when they actually hit something
		if ( (bg_itemlist[pm->ps->weapon].giTag == WP_GAUNTLET || bg_itemlist[pm->ps->weapon].giType != IT_WEAPON) && !pm->gauntletHit ) 
		{
			pm->ps->weaponTime = 0;
			pm->ps->weaponstate = WEAPON_READY;
			return;
		}		
		PM_StartTorsoAnim( TORSO_ATTACK2 );
		if(bg_itemlist[pm->ps->weapon].giType == IT_WEAPON && bg_itemlist[pm->ps->weapon].giTag == WP_SWORD)
		{
			switch(rand() & 3)
			{
			case 0: 
			case 1:
				PM_StartTorsoAnim( TORSO_RAISE );
				break;
			case 2:
				PM_StartTorsoAnim( TORSO_DROP );
				break;
			case 3:
				break; // stay with TORSO_ATTACK2
			}
		}
	} 
	else 
	{
		PM_StartTorsoAnim( TORSO_ATTACK );
	}

	pm->ps->weaponstate = WEAPON_FIRING;
	addTime = 50;

	// check for out of ammo
	if(!pm->ps->weapon)
	{ // we can ignore no ammo for the fist

	}
	else if(bg_itemlist[pm->ps->weapon].giType == IT_WEAPON && bg_itemlist[pm->ps->weapon].giTag == WP_GAUNTLET && pm->ps->ammo[AMMO_CURRENT] == 1)
	{ // warn them that they just ran out
		PM_AddEvent( EV_NOAMMO );
	}
	else if ( !pm->ps->ammo[ AMMO_CURRENT ] && !pm->gauntletHit ) 
	{
		PM_AddEvent( EV_NOAMMO );
		pm->ps->weaponTime += 500;
		return;
	}
	else
	{
		
		if(bg_itemlist[pm->ps->weapon].giType == IT_WEAPON && pm->ps->ammo[AMMO_CURRENT] > 0)
		{
			PM_AddEvent( EV_FIRE_WEAPON );
			pm->ps->ammo[AMMO_CURRENT]--;
			switch( bg_itemlist[pm->ps->weapon].giTag ) {
			default:			
			case WP_BALL:
				addTime = 600;
				break;
			case WP_NONE:
				addTime = 350;
				break;
			case WP_GAUNTLET:
				addTime = 400;
				break;
			case WP_LIGHTNING:
				addTime = 50;
				break;
			case WP_FLAMETHROWER:
				addTime = 75;
				break;
			case WP_SHOTGUN:
				if(pm->ps->ammo[AMMO_CURRENT] & 1)
					addTime = 100; // fire second barrel next frame
				else
					addTime = 900;
				break;
			case WP_SINGLE_SHOTGUN:
				addTime = 1000;
				break;
			case WP_MACHINEGUN:
				addTime = 160;
				break;
			case WP_CHAINGUN:
				addTime = 50;
				break;
			case WP_GRENADE_LAUNCHER:
				addTime = 800;
				break;
			case WP_ROCKET_LAUNCHER:
				addTime = 800;
				break;
			case WP_PLASMAGUN:
				addTime = 100;
				break;
			case WP_RAILGUN:
				addTime = 1500;
				break;
			case WP_SWORD:
				switch(pm->ps->torsoAnim & ~ANIM_TOGGLEBIT)
				{
				case TORSO_ATTACK:
				case TORSO_ATTACK2:
					addTime = 500;
					break;
				case TORSO_RAISE:
					addTime = 1000;
					break;
				case TORSO_DROP:
					addTime = 750;
					break;
				}
			}
		}
		else if(bg_itemlist[pm->ps->weapon].giType == IT_HEALTH && pm->ps->ammo[AMMO_CURRENT] > 0)
		{
			int addhealth = 100;
			float weight = pm->ps->stats[STAT_LEG_DAMAGE] + pm->ps->stats[STAT_HEAD_DAMAGE] + pm->ps->stats[STAT_BLEEDING] + (MAX_HEALTH - pm->ps->stats[STAT_LEG_DAMAGE]);
			if(bg_itemlist[pm->ps->weapon].giTag == PW_BERZERK)
			{
				PM_AddEvent(EV_USE_BERZERK);	
			}
			else if(weight < 100)
			{
				pm->ps->stats[STAT_LEG_DAMAGE] = 0;
				pm->ps->stats[STAT_HEAD_DAMAGE] = 0;
				pm->ps->stats[STAT_BLEEDING] = 0;
				pm->ps->stats[STAT_HEALTH] = MAX_HEALTH;
			}
			else
			{
				pm->ps->stats[STAT_LEG_DAMAGE] *= (1 -  100 / weight);
				pm->ps->stats[STAT_BLEEDING] *= (1 -  100 / weight);
				pm->ps->stats[STAT_HEAD_DAMAGE] *= (1 -  100 / weight);
				pm->ps->stats[STAT_HEALTH] += MAX_HEALTH * 10 / weight;
				if(pm->ps->stats[STAT_HEALTH] > MAX_HEALTH)
				{
					pm->ps->stats[STAT_BLEEDING] -= pm->ps->stats[STAT_HEALTH] - MAX_HEALTH;
					if(pm->ps->stats[STAT_BLEEDING] < 0)
					{
						pm->ps->stats[STAT_LEG_DAMAGE] += pm->ps->stats[STAT_BLEEDING];
						if(pm->ps->stats[STAT_LEG_DAMAGE] < 0)
						{
							pm->ps->stats[STAT_HEAD_DAMAGE] += pm->ps->stats[STAT_LEG_DAMAGE];
							if(pm->ps->stats[STAT_HEAD_DAMAGE] < 0)
								pm->ps->stats[STAT_HEAD_DAMAGE] = 0;
							pm->ps->stats[STAT_LEG_DAMAGE] = 0;
						}
						pm->ps->stats[STAT_BLEEDING] = 0;
					}
					pm->ps->stats[STAT_HEALTH] = MAX_HEALTH;
				}
			}
			pm->ps->ammo[AMMO_CURRENT]--;

			addTime = 450;
		}

		else if(bg_itemlist[pm->ps->weapon].giType == IT_AMMO && bg_itemlist[pm->ps->weapon].giTag == AMMO_GRENADES && pm->ps->ammo[AMMO_CURRENT] == 1)
		{
			PM_AddEvent( EV_NOAMMO );
		}
		else
		{
			addTime = 350;
			addTime += bg_itemlist[pm->ps->weapon].mass * 10;
		}
	}
	if ( pm->ps->powerups[PW_HASTE] ) {
		addTime /= g_HasteFactor.value;
	}

	pm->ps->weaponTime += addTime;
}

/*
================
PM_Animate
================
*/
static void PM_Animate( void ) {
	if ( pm->cmd.buttons & BUTTON_GESTURE ) {
		if ( pm->ps->torsoTimer == 0 ) {
			PM_StartTorsoAnim( TORSO_GESTURE );
			pm->ps->torsoTimer = TIMER_GESTURE;
			PM_AddEvent( EV_TAUNT );
		}
	}
}


/*
================
PM_DropTimers
================
*/
static void PM_DropTimers( void ) {
	// drop misc timing counter
	if ( pm->ps->pm_time ) {
		if ( pml.msec >= pm->ps->pm_time ) {
			pm->ps->pm_flags &= ~PMF_ALL_TIMES;
			pm->ps->pm_time = 0;
		} else {
			pm->ps->pm_time -= pml.msec;
		}
	}

	// drop animation counter
	if ( pm->ps->legsTimer > 0 ) {
		pm->ps->legsTimer -= pml.msec;
		if ( pm->ps->legsTimer < 0 ) {
			pm->ps->legsTimer = 0;
		}
	}

	if ( pm->ps->torsoTimer > 0 ) {
		pm->ps->torsoTimer -= pml.msec;
		if ( pm->ps->torsoTimer < 0 ) {
			pm->ps->torsoTimer = 0;
		}
	}
}

/*
================
PM_UpdateViewAngles

This can be used as another entry point when only the viewangles
are being updated isntead of a full move
================
*/
void PM_UpdateViewAngles( playerState_t *ps, const usercmd_t *cmd ) {
	short		temp;
	int		i;

	if ( ps->pm_type == PM_INTERMISSION ) {
		return;		// no view changes at all
	}

	if ( ps->pm_type != PM_SPECTATOR && ps->stats[STAT_HEALTH] <= 0 ) {
		return;		// no view changes at all
	}

	// circularly clamp the angles with deltas
	for (i=0 ; i<3 ; i++) {		
		temp = cmd->angles[i] + ps->delta_angles[i];
		
		if ( (i == PITCH) )
		{
			// don't let the player look up or down more than 90 degrees
			if(ps->groundEntityNum == ENTITYNUM_NONE)
			{
				if((ps->legsAnim & ~ANIM_TOGGLEBIT) == LEGS_JUMP && (temp > 16383 || temp < -16383) )
				{
					ps->delta_angles[i] += 64 * pml.msec;
					temp += 64 * pml.msec;
				}
				else if((ps->legsAnim & ~ANIM_TOGGLEBIT) == LEGS_JUMPB && (temp > 16383 || temp < -16383) )
				{
					ps->delta_angles[i] -= 64 * pml.msec;
					temp -= 64 * pml.msec;
				}
			}
			
			else if ( temp > 16383) 
			{
				ps->delta_angles[i] = 16383 - cmd->angles[i];
				temp = 16383;
			}
			else if ( temp < -16383) 
			{
				ps->delta_angles[i] = -16383 - cmd->angles[i];
				temp = -16383;
			}
		}
		else if ( (i == YAW) )
		{
			if( fabs(ps->viewangles[PITCH]) > 16383 )
			{
				float diff = cmd->angles[i] - ps->viewangles[i];
				
				temp -= diff * 2;
				ps->delta_angles[i] -= diff * 2;
			}
		}
		ps->viewangles[i] = SHORT2ANGLE(temp);
	}

}


/*
================
PmoveSingle

================
*/
void PmoveSingle (pmove_t *pmove) {
	pm = pmove;
	// this counter lets us debug movement problems with a journal
	// by setting a conditional breakpoint fot the previous frame
	c_pmove++;

	// clear results
	pm->numtouch = 0;
	pm->watertype = 0;
	pm->waterlevel = 0;

#if 0 // Hentai - I don't get why we'd want this.
	if ( pm->ps->stats[STAT_HEALTH] <= 0 ) {
		pm->tracemask &= ~CONTENTS_BODY;	// corpses can fly through bodies
	}
#endif

	
	// make sure walking button is clear if they are running, to avoid
	// proxy no-footsteps cheats
	if ( abs( pm->cmd.forwardmove ) > 64 || abs( pm->cmd.rightmove ) > 64 || (pm->ps->stats[STAT_LEG_DAMAGE] * 2 > MAX_HEALTH)) {
		pm->cmd.buttons &= ~BUTTON_WALKING;
	}

	// set the talk balloon flag
	if ( pm->cmd.buttons & BUTTON_TALK ) {
		pm->ps->eFlags |= EF_TALK;
	} else {
		pm->ps->eFlags &= ~EF_TALK;
	}

	// set the firing flag for continuous beam weapons
	if ( !(pm->ps->pm_flags & PMF_RESPAWNED) && pm->ps->pm_type != PM_INTERMISSION
		&& ( pm->cmd.buttons & BUTTON_ATTACK ) && pm->ps->ammo[AMMO_CURRENT] ) {
		pm->ps->eFlags |= EF_FIRING;
	} else {
		pm->ps->eFlags &= ~EF_FIRING;
	}

	// clear the respawned flag if attack and use are cleared
	if ( pm->ps->stats[STAT_HEALTH] > 0 && 
		!( pm->cmd.buttons & (BUTTON_ATTACK | BUTTON_USE_HOLDABLE) ) ) {
		pm->ps->pm_flags &= ~PMF_RESPAWNED;
	}

	// if talk button is down, dissallow all other input
	// this is to prevent any possible intercept proxy from
	// adding fake talk balloons
	if ( pmove->cmd.buttons & BUTTON_TALK ) {
		pmove->cmd.buttons = 0;
		pmove->cmd.forwardmove = 0;
		pmove->cmd.rightmove = 0;
		pmove->cmd.upmove = 0;
	}

	// clear all pmove local vars
	memset (&pml, 0, sizeof(pml));

	// determine the time
	pml.msec = pmove->cmd.serverTime - pm->ps->commandTime;
	if ( pml.msec < 1 ) {
		pml.msec = 1;
	} else if ( pml.msec > 200 ) {
		pml.msec = 200;
	}
	pm->ps->commandTime = pmove->cmd.serverTime;

	// save old org in case we get stuck
	VectorCopy (pm->ps->origin, pml.previous_origin);

	// save old velocity for crashlanding
	VectorCopy (pm->ps->velocity, pml.previous_velocity);

	pml.frametime = pml.msec * 0.001;

	// update the viewangles
	PM_UpdateViewAngles( pm->ps, &pm->cmd );

	AngleVectors (pm->ps->viewangles, pml.forward, pml.right, pml.up);

	if ( pm->cmd.upmove < 10 ) {
		// not holding jump
		pm->ps->pm_flags &= ~PMF_JUMP_HELD;
	}

	// decide if backpedaling animations should be used
	if ( pm->cmd.forwardmove < 0 ) {
		pm->ps->pm_flags |= PMF_BACKWARDS_RUN;
	} else if ( pm->cmd.forwardmove > 0 || ( pm->cmd.forwardmove == 0 && pm->cmd.rightmove ) ) {
		pm->ps->pm_flags &= ~PMF_BACKWARDS_RUN;
	}

	if ( pm->ps->pm_type >= PM_DEAD ) {
		pm->cmd.forwardmove = 0;
		pm->cmd.rightmove = 0;
		pm->cmd.upmove = 0;
	}

	if ( pm->ps->pm_type == PM_SPECTATOR ) {
		PM_CheckDuck ();
		PM_FlyMove ();
		PM_DropTimers ();
		return;
	}

	if ( pm->ps->pm_type == PM_NOCLIP ) {
		PM_NoclipMove ();
		PM_DropTimers ();
		return;
	}

	if (pm->ps->pm_type == PM_FREEZE) {
		return;		// no movement at all
	}

	if ( pm->ps->pm_type == PM_INTERMISSION ) {
		return;		// no movement at all
	}

	// set watertype, and waterlevel
	PM_SetWaterLevel();
	pml.previous_waterlevel = pmove->waterlevel;

	// set mins, maxs, and viewheight
	PM_CheckDuck ();

	// set groundentity
	PM_GroundTrace();
	
	
	if ( pm->ps->pm_type == PM_DEAD ) {
		PM_DeadMove ();
	}

	PM_DropTimers();
	//
	if ( pm->ps->powerups[PW_FLIGHT] ) {
		// flight powerup doesn't allow jump and has different friction
		PM_FlyMove();
	} else if (pm->ps->pm_flags & PMF_GRAPPLE_PULL) {
		PM_GrappleMove();
		// We can wiggle a bit
		PM_AirMove();
	} else if (pm->ps->pm_flags & PMF_TIME_WATERJUMP) {
		PM_WaterJumpMove();
	} else if ( pm->waterlevel > 1 ) {
		// swimming
		PM_WaterMove();
	
	}
	else if((pm->ps->stats[STAT_EXTENDED_INFO] & EXT_WALLCLIMBER) && (pm->cmd.upmove < 0))
	{

		if(pml.groundPlane)
		{
			PM_CrawlMove();
			
		}
		else
			PM_AirMove();
	} else if ( pml.walking ) {
		// walking on ground
		PM_WalkMove();
	} else {
		// airborne
		PM_AirMove();
	}

	PM_Animate();

	// set groundentity, watertype, and waterlevel
	PM_GroundTrace();
	PM_SetWaterLevel();

	// weapons
	PM_Weapon();

	// torso animation
	PM_TorsoAnimation();

	// footstep events / legs animations
	PM_Footsteps();

	// entering / leaving water splashes
	PM_WaterEvents();

	// snap some parts of playerstate to save network bandwidth
	SnapVector( pm->ps->velocity );
}


/*
================
Pmove

Can be called by either the server or the client
================
*/
void Pmove (pmove_t *pmove) {
	int			finalTime;
	pm_duckScale = 0.25;
	finalTime = pmove->cmd.serverTime;


	if ( finalTime < pmove->ps->commandTime ) {
		return;	// should not happen
	}

	if ( finalTime > pmove->ps->commandTime + 1000 ) {
		pmove->ps->commandTime = finalTime - 1000;
	}

	// chop the move up if it is too long, to prevent framerate
	// dependent behavior
	while ( pmove->ps->commandTime != finalTime ) {
		int		msec;

		msec = finalTime - pmove->ps->commandTime;

		if ( msec > 66 ) {
			msec = 66;
		}
		pmove->cmd.serverTime = pmove->ps->commandTime + msec;
		PmoveSingle( pmove );
	}

}
